from http.client import NotConnected
from koil import koil
from rath.errors import NotConnectedError
from rath.links.base import TerminatingLink
import asyncio
from typing import Dict, Any, Optional, List, Union, Callable, Awaitable
from rath.operation import GraphQLResult, opify
from contextvars import ContextVar


class Rath:
    def __init__(
        self,
        link: TerminatingLink,
        register=False,
        autoconnect=True,
    ) -> None:
        self.link = link
        self.autoconnect = autoconnect
        self.link(self)
        self.connected = False

        if register:
            set_current_rath(self)

    async def aconnect(self):
        await self.link.aconnect()
        self.connected = True

    def connect(self):
        return koil(self.aconnect())

    async def asubscribe(self):
        pass

    async def aexecute(
        self,
        query: str,
        variables: Dict[str, Any] = None,
        headers: Dict[str, Any] = {},
        operation_name="",
        **kwargs,
    ) -> GraphQLResult:

        if not self.connected:
            if not self.autoconnect:
                raise NotConnectedError(
                    "Rath is not connected and autoconnect is set to false. Please connect first or set autoconnect to true."
                )

            await self.aconnect()

        op = opify(query, variables, headers, operation_name, **kwargs)

        for parser in self.parsers:
            op = await parser.aparse(op)

        return await self.link.aquery(
            op,
        )

    def execute(
        self,
        query: str,
        variables: Dict[str, Any] = None,
        headers: Dict[str, Any] = {},
        operation_name="",
        **kwargs,
    ) -> GraphQLResult:
        op = opify(query, variables, headers, operation_name, **kwargs)

        if not self.connected:
            if not self.autoconnect:
                raise NotConnectedError(
                    "Rath is not connected and autoconnect is set to false. Please connect first or set autoconnect to true."
                )
            koil(self.aconnect())

        return self.link.query(
            op,
        )

    def subscribe(
        self,
        query: str,
        variables: Dict[str, Any] = None,
        headers: Dict[str, Any] = {},
        operation_name="",
        **kwargs,
    ) -> GraphQLResult:

        op = opify(query, variables, headers, operation_name, **kwargs)
        for res in self.link.subscribe(op):
            yield res

    async def asubscribe(
        self,
        query: str,
        variables: Dict[str, Any] = None,
        headers: Dict[str, Any] = {},
        operation_name="",
        **kwargs,
    ) -> GraphQLResult:
        op = opify(query, variables, headers, operation_name, **kwargs)
        async for res in self.link.asubscribe(op):
            yield res


CURRENT_RATH = None


def get_current_rath(**kwargs):
    global CURRENT_RATH
    assert CURRENT_RATH is not None, "No current rath set"
    return CURRENT_RATH


def set_current_rath(rath):
    global CURRENT_RATH
    CURRENT_RATH = rath








from pydantic import BaseModel
from herre.herre import Herre, get_current_herre
from rath.links.auth import AuthTokenLink



class HerreAuthTokenLink(AuthTokenLink):


    def __init__(self, *args, herre: Herre = None, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.herre = herre or get_current_herre()
        self.config = None

    async def aconnect(self) -> None:
        self.token_loader = self.herre.aget_token
        return await super().aconnect()



from typing import Any, Coroutine
from rath.links.base import AsyncContinuationLink
from rath.operation import GraphQLResult, Operation
from rath.links.errors import AuthenticationError


async def fake_loader():
    raise Exception("No Token loader specified")


class AuthTokenLink(AsyncContinuationLink):
    def __init__(
        self,
        token_loader: Coroutine[Any, Any, str] = fake_loader,
        maximum_refresh_attempts=3,
        load_token_on_connect=True,
    ) -> None:
        self.token_loader = token_loader
        self.load_token_on_connect = load_token_on_connect
        self.maximum_refresh_attempts = maximum_refresh_attempts
        self.token = None

    async def reload_token(self) -> None:
        self.token = await self.token_loader()
        return self.token

    async def aquery(self, operation: Operation, retry=0) -> GraphQLResult:

        operation.context.headers["Authorization"] = f"Bearer {self.token}"
        try:
            return await self.next.aquery(operation)
        except AuthenticationError as e:
            if retry >= self.maximum_refresh_attempts:
                raise AuthenticationError("Maximum refresh attempts reached") from e

            self.token = await self.reload_token()
            return await self.aquery(operation, retry=retry + 1)

    async def asubscribe(self, operation: Operation, retry=0) -> GraphQLResult:
        if retry >= self.maximum_refresh_attempts:
            raise AuthenticationError("Maximum refresh attempts reached")

        operation.context.headers["Authorization"] = f"Bearer {self.token}"
        try:
            async for result in self.next.asubscribe(operation):
                yield result

        except AuthenticationError as e:
            async for result in self.asubscribe(operation, retry=retry + 1):
                yield result

    async def aconnect(self) -> None:
        if self.load_token_on_connect:
            self.token = await self.token_loader()



from dataclasses import dataclass, field
from typing import Optional, Dict, Any
from graphql.language import OperationDefinitionNode, parse


@dataclass
class Context:
    headers: Optional[Dict[str, str]] = field(default_factory=dict)
    files: Optional[Dict[str, Any]] = field(default_factory=dict)
    kwargs: Optional[Dict[str, Any]] = field(default_factory=dict)


@dataclass
class Extensions:
    pollInterval: Optional[int] = None
    maxPolls: Optional[int] = None


@dataclass
class Operation:
    node: OperationDefinitionNode
    document: str
    variables: Dict[str, Any]
    operation_name: str
    extensions: Extensions
    context: Context


@dataclass
class GraphQLResult:
    data: Dict[str, Any]


class GraphQLException(Exception):
    pass


def opify(
    query: str,
    variables: Dict[str, Any] = None,
    headers: Dict[str, Any] = {},
    operation_name: str = None,
    **kwargs,
) -> Operation:

    document = parse(query)

    op = [o for o in document.definitions if isinstance(o, OperationDefinitionNode)][0]

    return Operation(
        node=op,
        document=query,
        variables=variables or {},
        operation_name=operation_name,
        extensions={},
        context=Context(headers=headers, kwargs=kwargs),
    )



from typing import Any, List, Optional
from pydantic.class_validators import validator
from pydantic.fields import Field
from pydantic.main import BaseModel
from rath.turms.utils import clean_dict

# Mixin CLass is not Possible because of BaseModel Metaclass


class GraphQLObjectError(Exception):
    pass


class GraphQLObject(BaseModel):
    def dict(self, *args, by_alias=True, **kwargs):

        return super().dict(
            *args,
            **{
                **kwargs,
                "by_alias": by_alias,
            },
        )

    async def to_variable(self):
        """WIll be called by the ward"""
        assert (
            self.id
        ), "Cannot convert an object to a variable if you didn't query its unique id"
        return self.id

    def __getattr__(self, attr):
        raise GraphQLObjectError(
            f"{attr} is not a valid attribute of {self.__class__.__name__}. Have you checked the schema?"
        )

    async def shrink(self):
        """WIll be called by the ward"""
        assert (
            self.id
        ), "Cannot convert an object to a variable if you didn't query its unique id"
        return self.id


class GraphQLInputObject(BaseModel):
    def dict(self, *args, by_alias=True, **kwargs):

        return super().dict(
            *args,
            **{
                **kwargs,
                "by_alias": by_alias,
            },
        )

    async def to_variable(self):
        """WIll be called by the ward"""
        dictionary = self.dict(exclude={"typename"})
        clean_dict(dictionary, lambda key, value: key == "__typename" or value is None)
        return dictionary



import asyncio
from concurrent.futures import ThreadPoolExecutor
from koil.loop import koil_gen
from rath.links.base import ContinuationLink
from rath.operation import GraphQLResult, Operation
from koil import koil


class SwitchAsyncLink(ContinuationLink):
    async def aquery(self, operation: Operation) -> GraphQLResult:
        return self.next.aquery(operation)

    async def asubscibe(self, operation: Operation) -> GraphQLResult:
        async for result in self.next.aconnect(operation):
            yield result

    def query(self, operation: Operation) -> GraphQLResult:
        return koil(self.next.aquery(operation))

    def subscribe(self, operation: Operation) -> GraphQLResult:
        for result in koil_gen(self.next.asubscribe(operation)):
            yield result


class SwitchSyncLink(ContinuationLink):
    def __init__(self, excecutor=None) -> None:
        self.excecutor = excecutor or ThreadPoolExecutor()
        super().__init__()

    async def aconnect(self) -> None:
        self.e = self.excecutor.__enter__()

    async def aquery(self, operation: Operation) -> GraphQLResult:
        return await asyncio.wrap_future(self.e.submit(self.next.query, operation))

    async def asubscibe(self, operation: Operation) -> GraphQLResult:
        raise NotImplementedError(
            "We need to fiqure this out yet. Normally a __next__ call here would be enough"
        )

    def query(self, operation: Operation) -> GraphQLResult:
        return self.next.query(operation)

    def subscribe(self, operation: Operation) -> GraphQLResult:
        for result in self.next.subscribe(operation):
            yield result



from pydantic.main import BaseModel, ModelMetaclass
from turms.types.operation import OperationMeta
from rath.rath import Rath, get_current_rath


class GraphQLOperation(BaseModel, metaclass=OperationMeta):
    @classmethod
    def execute(cls, variables, rath: Rath = None):
        rath = rath or get_current_rath()
        return cls(**rath.execute(cls.get_meta().document, variables).data)

    @classmethod
    async def aexecute(cls, variables, rath: Rath = None):
        rath = rath or get_current_rath()
        return cls(**(await rath.aexecute(cls.get_meta().document, variables)).data)

    class Meta:
        abstract = True


class GraphQLQuery(GraphQLOperation):
    class Meta:
        abstract = True


class GraphQLMutation(GraphQLOperation):
    class Meta:
        abstract = True


class GraphQLSubscription(GraphQLOperation):
    class Meta:
        abstract = True



from rath.operation import Operation
from rath.links.parsing import ParsingLink


class ShrinkingLink(ParsingLink):
    def __init__(self, file_path: str) -> None:
        super().__init__()
        self.file_path = file_path

    def parse(self, operation: Operation) -> Operation:
        shrinked_variables = {
            key: var.shrink()
            for key, var in operation.variables.items()
            if hasattr(var, "shrink")
        }
        operation.variables.update(shrinked_variables)
        return operation

    async def aparse(self, operation: Operation) -> Operation:
        shrinked_variables = {
            key: await var.ashrink()
            for key, var in operation.variables.items()
            if hasattr(var, "shrink")
        }
        operation.variables.update(shrinked_variables)
        return operation



from rath.operation import GraphQLResult, Operation


class Link:
    async def aconnect(self) -> None:
        pass

    async def aquery(self, operation: Operation) -> GraphQLResult:
        raise NotImplementedError(
            f"Please overwrite the aquery method in {self.__class__.__name__}"
        )

    async def asubscribe(self, operation: Operation) -> GraphQLResult:
        raise NotImplementedError(
            f"Please overwrite the asubscribe method in {self.__class__.__name__}"
        )

    def query(self, operation: Operation) -> GraphQLResult:
        raise NotImplementedError(
            f"Please overwrite the query method in {self.__class__.__name__}"
        )

    def subscribe(self, operation: Operation) -> GraphQLResult:
        raise NotImplementedError(
            f"Please overwrite the subscribe method in {self.__class__.__name__}"
        )


class TerminatingLink(Link):
    def __call__(self, rath):
        self.rath = rath
        return self


class AsyncTerminatingLink(TerminatingLink):
    def query(self, operation: Operation) -> GraphQLResult:
        raise NotImplementedError(
            "Async Terminating link does not support syncrhonous queries. Please compose together with a context switch link"
        )

    def subscribe(self, operation: Operation) -> GraphQLResult:
        raise NotImplementedError(
            "Async Terminating link does not support syncrhonous queries. Please compose together with a context switch link"
        )


class SyncTerminatingLink(TerminatingLink):
    async def aquery(self, operation: Operation) -> GraphQLResult:
        raise NotImplementedError(
            "Sync Terminating link does not support asynchronous queries. Please compose together with a context switch link"
        )

    async def asubscribe(self, operation: Operation) -> GraphQLResult:
        raise NotImplementedError(
            "Sync Terminating link does not support asynchronous queries. Please compose together with a context switch link"
        )


class ContinuationLink(Link):
    def __call__(self, rath, next: Link):
        self.rath = rath
        self.next = next
        return self


class AsyncContinuationLink(ContinuationLink):
    def query(self, operation: Operation) -> GraphQLResult:
        raise NotImplementedError(
            "Async Terminating link does not support syncrhonous queries. Please compose together with a context switch link"
        )

    def subscribe(self, operation: Operation) -> GraphQLResult:
        raise NotImplementedError(
            "Async Terminating link does not support syncrhonous queries. Please compose together with a context switch link"
        )


class SyncContinuationLink(ContinuationLink):
    async def aquery(self, operation: Operation) -> GraphQLResult:
        raise NotImplementedError(
            "Sync Terminating link does not support asynchronous queries. Please compose together with a context switch link"
        )

    async def asubscribe(self, operation: Operation) -> GraphQLResult:
        raise NotImplementedError(
            "Sync Terminating link does not support asynchronous queries. Please compose together with a context switch link"
        )



from rath.links.base import ContinuationLink
from rath.operation import Operation


class ParsingLink(ContinuationLink):
    def __init__(self, file_path: str) -> None:
        super().__init__()
        self.file_path = file_path

    def parse(self, operation: Operation) -> Operation:
        raise NotImplementedError("Please implement this method")

    def aparse(self, operation: Operation) -> Operation:
        raise NotImplementedError("Please implement this method")

    async def aquery(self, operation: Operation) -> Operation:
        operation = self.aparse(operation)
        return await self.next.aquery(operation)

    async def asubscribe(self, operation: Operation) -> Operation:
        operation = self.aparse(operation)
        async for result in self.next.asubscribe(operation):
            yield result

    def query(self, operation: Operation) -> Operation:
        operation = self.parse(operation)
        return self.next.query(operation)

    def subscribe(self, operation: Operation) -> Operation:
        operation = self.parse(operation)
        for result in self.next.subscribe(operation):
            yield result



from pydantic import BaseModel
from fakts.fakts import Fakts, get_current_fakts
from rath.links.aiohttp import AIOHttpLink


class FaktsAioHttpLinkConfig(BaseModel):
    endpoint_url: str


class FaktsAioHttpLink(AIOHttpLink):


    def __init__(self, *args, fakts: Fakts = None, fakts_key="rath", **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.fakts = fakts or get_current_fakts()
        self.fakts_key = fakts_key
        self.config = None


    def configure(self, fakts: FaktsAioHttpLinkConfig):
        self.url = fakts.endpoint_url


    async def aconnect(self) -> None:
        if not self.config:
            self.fakts = await self.fakts.aget(self.fakts_key)
            self.configure(FaktsAioHttpLinkConfig(**self.fakts))

        return await super().aconnect()



import asyncio
from typing import Dict, Type
from herre import get_current_herre
from rath.rath import Rath


class NoRathRegistered(Exception):
    pass


class RathRegistry:
    def __init__(self) -> None:
        self.domainRathClassMap: Dict[str, Type[Rath]] = {}
        self.domainRathInstanceMap: Dict[str, Rath] = {}

    def register_rath(self, domain, wardClass: Type[Rath]):
        assert (
            domain not in self.domainRathClassMap
        ), "We cannot register another Ward for this domain"
        self.domainRathClassMap[domain] = wardClass

    def get_rath_class(self, domain) -> Rath:
        try:
            return self.domainRathClassMap[domain]
        except KeyError as e:
            raise NoRathRegistered(f"No Rath registered for domain {domain}") from e

    def get_rath_instance(self, domain) -> Rath:
        if domain in self.domainRathInstanceMap:
            return self.domainRathInstanceMap[domain]
        self.domainRathInstanceMap[domain] = self.domainRathClassMap[domain]()
        return self.domainRathInstanceMap[domain]

    # Handling Connection

    async def connect_all(self):
        unconnected_wards = [
            ward
            for key, ward in self.domainRathInstanceMap.items()
            if not ward.connected
        ]
        await asyncio.gather(*[ward._connect() for ward in unconnected_wards])

    async def unconnect_all(self):
        connected_wards = [
            ward for key, ward in self.domainRathInstanceMap.items() if ward.connected
        ]
        await asyncio.gather(*[ward._disconnect() for ward in connected_wards])


RATH_REGISTRY = None


def get_rath_registry():
    global RATH_REGISTRY
    if not RATH_REGISTRY:
        RATH_REGISTRY = RathRegistry()
    return RATH_REGISTRY



from asyncio.tasks import create_task
from typing import Dict
from rath.links.errors import TransportError
from graphql import OperationType
import websockets
import json
import asyncio
from websockets.exceptions import (
    ConnectionClosed,
    ConnectionClosedError,
    ConnectionClosedOK,
)
import logging
import uuid

from rath.operation import GraphQLResult, Operation
from rath.links.base import AsyncTerminatingLink, TerminatingLink, Transport

logger = logging.getLogger(__name__)


GQL_WS_SUBPROTOCOL = "graphql-ws"

# all the message types
GQL_CONNECTION_INIT = "connection_init"
GQL_START = "start"
GQL_STOP = "stop"
GQL_CONNECTION_TERMINATE = "connection_terminate"
GQL_CONNECTION_ERROR = "connection_error"
GQL_CONNECTION_ACK = "connection_ack"
GQL_DATA = "data"
GQL_ERROR = "error"
GQL_COMPLETE = "complete"
GQL_CONNECTION_KEEP_ALIVE = "ka"


class CorrectableConnectionFail(TransportError):
    pass


class DefiniteConnectionFail(TransportError):
    pass


class InvalidPayload(TransportError):
    pass


class WebSocketLink(AsyncTerminatingLink):
    def __init__(
        self, url="", allow_reconnect=True, time_between_retries=1, retries=3
    ) -> None:
        self.connection_initialized = False
        self.ongoing_subscriptions: Dict[str, asyncio.Queue] = {}
        self.url = url
        self.retries = retries
        self.allow_reconnect = allow_reconnect
        self.time_between_retries = time_between_retries
        pass

    async def aforward(self, message):
        await self.send_queue.put(message)

    async def aconnect(self):
        self.send_queue = asyncio.Queue()
        self.connection_task = create_task(self.websocket_loop())

    async def adisconnect(self):
        self.connection_task.cancel()

        try:
            await self.connection_task
        except asyncio.CancelledError:
            logger.info(f"Websocket Transport {self} succesfully disconnected")

    async def websocket_loop(self, retry=0):
        send_task = None
        receive_task = None
        self.connection_initialized = False
        try:
            try:
                async with websockets.connect(
                    self.url,
                    subprotocols=[GQL_WS_SUBPROTOCOL],
                ) as client:

                    send_task = create_task(self.sending(client))
                    receive_task = create_task(self.receiving(client))

                    self.connection_alive = True
                    self.connection_dead = False
                    done, pending = await asyncio.wait(
                        [send_task, receive_task],
                        return_when=asyncio.FIRST_EXCEPTION,
                    )
                    self.connection_alive = True

                    for task in pending:
                        task.cancel()

                    for task in done:
                        raise task.exception()

            except ConnectionClosedError as e:
                logger.exception(e)
                raise CorrectableConnectionFail from e

            except Exception as e:
                print(e)
                raise CorrectableConnectionFail from e

        except CorrectableConnectionFail as e:
            logger.info(f"Trying to Recover from Exception {e}")
            if retry > self.retries or not self.allow_reconnect:
                raise DefiniteConnectionFail("Exceeded Number of Retries")

            await asyncio.sleep(self.time_between_retries)
            logger.info(f"Retrying to connect")
            await self.websocket_loop(retry=retry + 1)

        except DefiniteConnectionFail as e:
            self.connection_dead = False
            raise e

        except asyncio.CancelledError as e:
            logger.info("Got Canceleld")
            if send_task and receive_task:
                send_task.cancel()
                receive_task.cancel()

            cancellation = await asyncio.gather(
                send_task, receive_task, return_exceptions=True
            )
            raise e

    async def sending(self, client, headers=None):
        payload = {"type": GQL_CONNECTION_INIT, "payload": {"headers": headers}}
        await client.send(json.dumps(payload))

        try:
            while True:
                message = await self.send_queue.get()
                logger.debug("GraphQL Websocket: >>>>>> " + message)
                await client.send(message)
                self.send_queue.task_done()
        except asyncio.CancelledError as e:
            logger.debug("Sending Task sucessfully Cancelled")

    async def receiving(self, client):
        try:
            async for message in client:
                logger.debug("Postman Websocket: <<<<<<< " + message)
                await self.broadcast(message)
        except asyncio.CancelledError as e:
            logger.debug("Receiving Task sucessfully Cancelled")

    async def broadcast(self, res):
        try:
            message = json.loads(res)
        except json.JSONDecodeError as err:
            logger.warning(
                "Ignoring. Server sent invalid JSON data: %s \n %s", res, err
            )

        type = message["type"]

        if type == GQL_CONNECTION_KEEP_ALIVE:
            return

        if type in [GQL_DATA, GQL_COMPLETE]:

            if "id" not in message:
                raise InvalidPayload(f"Protocol Violation. Expected 'id' in {message}")

            id = message["id"]
            assert (
                id in self.ongoing_subscriptions
            ), "Received Result for subscription that is no longer or was never active"
            await self.ongoing_subscriptions[id].put(message)

        print(message)

    async def asubscribe(self, operation: Operation):

        assert (
            operation.node.operation == OperationType.SUBSCRIPTION
        ), "Operation is not a subscription"
        assert (
            operation.context.files is None
        ), "We cannot send files through websockets"

        id = str(uuid.uuid4())
        subscribe_queue = asyncio.Queue()
        self.ongoing_subscriptions[id] = subscribe_queue

        payload = {
            "headers": operation.context.headers,
            "query": operation.document,
            "variables": operation.variables,
        }
        frame = {"id": id, "type": GQL_START, "payload": payload}
        await self.aforward(json.dumps(frame))

        while True:
            answer = await subscribe_queue.get()

            if answer["type"] == GQL_DATA:
                payload = answer["payload"]

                if "data" in payload:
                    yield GraphQLResult(data=payload["data"])

            if answer["type"] == GQL_COMPLETE:
                print("Subcription done")
                return

    async def aquery(self, operation: Operation):

        assert (
            operation.node.operation != OperationType.SUBSCRIPTION
        ), "Operation is a subscription. Only queries are allowed on 'aquery'"
        assert (
            operation.context.files is None
        ), "We cannot send files through websockets"

        id = str(uuid.uuid4())
        subscribe_queue = asyncio.Queue()
        self.ongoing_subscriptions[id] = subscribe_queue

        payload = {
            "headers": operation.context.headers,
            "query": operation.document,
            "variables": operation.variables,
        }
        frame = {"id": id, "type": GQL_START, "payload": payload}
        await self.aforward(json.dumps(frame))

        while True:
            answer = await subscribe_queue.get()

            if answer["type"] == GQL_DATA:
                payload = answer["payload"]

                if "data" in payload:
                    subscribe_queue.task_done()
                    del self.ongoing_subscriptions[id]
                    return GraphQLResult(data=payload["data"])

            if answer["type"] == GQL_COMPLETE:
                raise InvalidPayload(
                    "Subcription done before yielding data. This shouldnt happen for queries"
                )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['rath', 'rath.fakts', 'rath.herre', 'rath.links', 'rath.turms']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.8.1,<4.0.0',
 'graphql-core>=3.2.0,<4.0.0',
 'koil>=0.1.59,<0.2.0',
 'pydantic>=1.9.0,<2.0.0']

setup_kwargs = {
    'name': 'rath',
    'version': '0.1.5',
    'description': 'aiohttp powered apollo like graphql client',
    'long_description': '# rath\n\n[![codecov](https://codecov.io/gh/jhnnsrs/rath/branch/master/graph/badge.svg?token=UGXEA2THBV)](https://codecov.io/gh/jhnnsrs/rath)\n[![PyPI version](https://badge.fury.io/py/rath.svg)](https://pypi.org/project/rath/)\n[![Maintenance](https://img.shields.io/badge/Maintained%3F-yes-green.svg)](https://pypi.org/project/rath/)\n![Maintainer](https://img.shields.io/badge/maintainer-jhnnsrs-blue)\n[![PyPI pyversions](https://img.shields.io/pypi/pyversions/rath.svg)](https://pypi.python.org/pypi/rath/)\n[![PyPI status](https://img.shields.io/pypi/status/rath.svg)](https://pypi.python.org/pypi/rath/)\n[![PyPI download month](https://img.shields.io/pypi/dm/rath.svg)](https://pypi.python.org/pypi/rath/)\n\n### DEVELOPMENT\n\n## Inspiration\n\nRath is like Apollo, but for python. It adheres to the design principle of Links and enables complex GraphQL\nsetups, like seperation of query and subscription endpoints, dynamic token loading, etc..\n\n## Installation\n\n```bash\npip install rath\n```\n\n## Usage Example\n\n```python\nfrom rath.links.auth import AuthTokenLink\nfrom rath.links.aiohttp import AioHttpLink\nfrom rath.links import compose, split\nfrom rath.gql import gql\n\nasync def aload_token():\n    return "SERVER_TOKEN"\n\n\nauth = AuthTokenLink(token_loader=aload_token)\nlink = AioHttpLink(url="https://api.spacex.land/graphql/")\n\n\nrath = Rath(links=compose(auth,link))\nrath.connect()\n\n\nquery = """query TestQuery {\n  capsules {\n    id\n    missions {\n      flight\n    }\n  }\n}\n"""\n\nresult = rath.execute(query)\n```\n\nThis example composes both the AuthToken and AioHttp link: During each query the Bearer headers are set to the retrieved token, on authentication fail (for example if Token Expired) the\nAuthToken automatically refetches the token and retries the query.\n\n## Async Usage\n\nRath is build with koil, for async/sync compatibility but also exposed a complete asynhronous api\n\n```python\nfrom rath.links.auth import AuthTokenLink\nfrom rath.links.aiohttp import AioHttpLink\nfrom rath.links import compose, split\nfrom rath.gql import gql\n\nasync def aload_token():\n    return "SERVER_TOKEN"\n\n\nauth = AuthTokenLink(token_loader=aload_token)\nlink = AioHttpLink(url="https://api.spacex.land/graphql/")\n\n\nasync def main():\n  rath = Rath(links=compose(auth,link))\n  await rath.aconnect()\n\n\n  query = """query TestQuery {\n    capsules {\n      id\n      missions {\n        flight\n      }\n    }\n  }\n  """\n\n  result = await rath.aexecute(query)\n\nasyncio.run(main())\n```\n\n## Included Links\n\n- Reconnecting WebsocketLink (untested)\n- AioHttpLink (supports multipart uploads)\n- SplitLink (allows to split the terminating link - Subscription into WebsocketLink, Query, Mutation into Aiohttp)\n- AuthTokenLink (Token insertion with automatic refresh)\n\n### Why Rath\n\nWell "apollo" is already taken as a name, and rath (according to wikipedia) is an etruscan deity identified with Apollo.\n\n## Rath + Turms\n\nRath works especially well with turms generated typed operations:\n\n```python\nimport asyncio\nfrom examples.api.schema import aget_capsules\nfrom rath.rath import Rath\nfrom rath.links.aiohttp import AIOHttpLink\nfrom rath.links.auth import AuthTokenLink\nfrom rath.links.compose import compose\n\n\nasync def token_loader():\n    return ""\n\n\nlink = compose(\n    AuthTokenLink(token_loader), AIOHttpLink("https://api.spacex.land/graphql/")\n)\n\n\nRATH = Rath(\n    link=link,\n    register=True, # allows global access (singleton-antipattern, but rath has no state)\n)\n\n\nasync def main():\n    capsules = await aget_capsules() # fully typed pydantic powered dataclasses generated through turms\n    print(capsules)\n\n\nasyncio.run(main())\n\n```\n\n## Examples\n\nThis github repository also contains an example client with a turms generated query with the public SpaceX api, as well as a sample of the generated api.\n\n## Parsers\n\nBesides links, there is also support for sequentially parsing of the Operation before it enters the asynchronous links (example in thread upload to\ns3fs), this is aided through parsers that provide both async and sync interfaces.\n',
    'author': 'jhnnsrs',
    'author_email': 'jhnnsrs@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



class LinkError(Exception):
    """Base class for all transport errors."""

    pass


class TerminatingLinkError(LinkError):
    """Raised when a terminating link is called."""

    pass


class ContinuationLinkError(LinkError):
    pass


class AuthenticationError(TerminatingLinkError):
    pass
